(*
Copyright 2012 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Check_Complexity
imports Reduction_Pair_Impl Reduction_Pair_Implementations
begin

subsection {*Proving complexity of TRSs*}

datatype ('f, 'v) complexity_proof =
  Rule_Shift_Complexity "'f redtriple_impl" "('f,'v)rules" "('f,'v)complexity_proof"
| RisEmpty_Complexity

fun
  check_complexity_proof ::
    "shows \<Rightarrow>
     ('tp, 'f::show, 'v :: show) tp_ops \<Rightarrow> 'tp \<Rightarrow>
     ('f,'v)complexity_measure \<Rightarrow> complexity_class \<Rightarrow>
     ('f, 'v) complexity_proof \<Rightarrow>
     shows check" 
where
  "check_complexity_proof i I tp cm cc (Rule_Shift_Complexity redp Rdelete prf) = debug (i []) ''Rule Removal Complexity'' (do {
         tp' \<leftarrow> rule_shift_complexity_tt I (get_redtriple redp) Rdelete cm cc tp
          <+? (\<lambda>s. i +@+ '': error when applying the rule shifting technique on ''
             +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ s);
         check_complexity_proof (i +@+ shows_string ''.1'') I tp' cm cc prf
           <+? (\<lambda>s. i +@+ '': error below the rule shifting processor'' +#+ shows_nl
             +@+ indent s)}
)" 
| "check_complexity_proof i I tp cm cc (RisEmpty_Complexity) = debug (i []) ''R is empty for complexity'' (do {
     check (tp_ops.R I tp = []) (i +@+ '': R is not empty in '' +#+ shows_nl +@+ shows_tp I tp) })"

lemma check_complexity_proof_sound: 
  assumes I: "tp_spec I" 
  and ok: "isOK (check_complexity_proof i I tp cm cc prf)"
  shows "deriv_bound_measure_class (rel_qrstep (tp_ops.qreltrs I tp)) cm cc"
proof -
  interpret tp_spec I by fact
  from ok show ?thesis
  proof (induct "prf" arbitrary: tp i cm cc)
    case RisEmpty_Complexity
    thus ?case by simp
  next
    case (Rule_Shift_Complexity rp Rdelete prof)
    note IH = Rule_Shift_Complexity 
    note res = IH(2)[simplified]    
    from res obtain tp' where rs: "rule_shift_complexity_tt I (get_redtriple rp) Rdelete cm cc tp = return tp'" by auto
    from res[unfolded rs, simplified] have rec: "isOK(check_complexity_proof (i +@+ shows_string ''.1'') I tp' cm cc prof)" by auto
    from IH(1)[OF this] have IH: "deriv_bound_measure_class (rel_qrstep (qreltrs tp')) cm cc" .
    from rule_shift_complexity_tt[OF get_redtriple, OF rs IH] show ?case .
  qed
qed

end