(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Koenigs_Lemma
imports
  Abstract_Rewriting
  "~~/src/HOL/Library/Infinite_Set"
begin

definition
  finitely_branching :: "'a rel \<Rightarrow> bool"
where
  "finitely_branching A \<equiv> \<forall>a. finite {b. (a, b) \<in> A}"

definition
  globally_finite :: "'a rel \<Rightarrow> bool"
where
  "globally_finite A \<equiv> \<forall>a. finite {b. (a, b) \<in> A^+}"

lemma split_successors:
  "{b. (a, b) \<in> A^+} = {b. (a, b) \<in> A} \<union> (\<Union>b\<in>{b. (a, b) \<in> A}. {c. (b, c) \<in> A^+})"
  (is "?succ = ?direct \<union> _")
proof
  show "?direct \<union> (\<Union>b\<in>?direct. {c. (b, c) \<in> A^+}) \<subseteq> ?succ" by auto
next
  show "?succ \<subseteq> ?direct \<union> (\<Union>b\<in>?direct. {c. (b, c) \<in> A^+})"
  proof
    fix b assume "b \<in> ?succ"
    hence "(a, b) \<in> A^+" by simp
    thus "b \<in> ?direct \<union> (\<Union>b\<in>?direct. {c. (b, c) \<in> A^+})" by (induct) auto
  qed
qed

lemma finitely_branching_SN_imp_globally_finite:
  assumes "finitely_branching A" and "SN A"
  shows "globally_finite A"
unfolding globally_finite_def
proof
  fix a
  from `SN A` show "finite {b. (a, b) \<in> A^+}"
  proof (induct)
    case (IH a)
    from `finitely_branching A` have "finite {b. (a, b) \<in> A}" by (simp add: finitely_branching_def)
    moreover from IH have "\<forall>b\<in>{b. (a, b) \<in> A}. finite {c. (b, c) \<in> A^+}" by auto
    ultimately show ?case unfolding split_successors[of a A] by auto
  qed
qed

lemma acyclic_globally_finite_imp_SN:
  assumes "acyclic A" and "globally_finite A"
  shows "SN A"
proof (rule ccontr)
  assume "\<not> SN A"
  then obtain a and S where "S 0 = a" and chain: "chain A S" by auto
  have "\<forall>i j. i \<noteq> j \<longrightarrow> S i \<noteq> S j"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then obtain i j where "i \<noteq> j" and "S i = S j" by auto
    hence "i < j \<or> j < i" by auto
    thus False
    proof
      assume "i < j"
      with chain_imp_trancl[OF chain this] and `S i = S j`
        have "\<not> acyclic A" unfolding acyclic_def by auto
      with assms show False by simp
    next
      assume "j < i"
      with chain_imp_trancl[OF chain this] and `S i = S j`
        have "\<not> acyclic A" unfolding acyclic_def by auto
      with assms show False by simp
    qed
  qed
  hence "inj S" unfolding inj_on_def by auto
  from range_inj_infinite[OF this] have "\<not> finite (range S)" .
  moreover have "finite (range S)"
  proof -
    have "range S \<subseteq> {a} \<union> {b. (a, b) \<in> A^+}"
    proof
      fix b assume "b \<in> range S"
      then obtain i where "S i = b" by auto
      show "b \<in> {a} \<union> {b. (a, b) \<in> A^+}"
      proof (cases i)
        case 0 thus ?thesis using `S i = b` and `S 0 = a` by auto
      next
        case (Suc j)
        hence "Suc j > 0" by simp
        from chain_imp_trancl[OF chain this]
          show ?thesis unfolding `S 0 = a` and Suc[symmetric] and `S i = b` by simp
      qed
    qed
    moreover from `globally_finite A`[unfolded globally_finite_def,THEN spec[of _ a]]
      have "finite ({a} \<union> {b. (a, b) \<in> A^+})" by auto
    ultimately show ?thesis by (rule finite_subset)
  qed
  ultimately show False ..
qed

lemma Koenigs_Lemma:
  assumes "acyclic A" and "finitely_branching A"
  shows "globally_finite A = SN A"
  using assms
    and finitely_branching_SN_imp_globally_finite[of A]
    and acyclic_globally_finite_imp_SN[of A] by blast

end
