(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Proof_Checker
imports
  Check_DPP
  Check_TP
  Check_CRP
  Check_Completion_Proof
  Check_Equational_Proof
  Check_Complexity
begin

datatype ('f, 'v) strategy =
  No_Strategy
| Innermost
| Innermost_Q "('f, 'v) term list"

fun strategy_to_Q :: "('f, 'v) strategy \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) term list" where
  "strategy_to_Q No_Strategy _ = []"
| "strategy_to_Q Innermost R = remdups (map fst R)"
| "strategy_to_Q (Innermost_Q Q) _ = Q"

fun rel_rules_of :: "('f,'v)rules option \<Rightarrow> ('f,'v)rules"
 where "rel_rules_of None = []"
     | "rel_rules_of (Some r) = r"

datatype ('f, 'v) input_problem =
  TRS "('f, 'v)strategy" "('f, 'v) rules" "('f, 'v) rules"
| DP "('f, 'v) rules" "('f, 'v) rules" "('f, 'v)strategy" "('f, 'v) rules" "('f, 'v) rules"
| EQS_TRS "('f, 'v) equation list" "('f, 'v) rules"
| EQS_EQ "('f, 'v) equation list" "('f, 'v) term \<times> ('f, 'v) term"
| Complexity_TRS "('f, 'v)strategy" "('f,'v)rules" "('f,'v)rules" "('f,'v)complexity_measure" complexity_class

datatype ('f, 'l, 'v) cert_problem = 
  TRS_Proof "(('f, 'l) lab, 'v) strategy"
            "(('f, 'l) lab, 'v) rules"
            "(('f, 'l) lab, 'v) rules option"
            "('f, 'l, 'v) trs_proof"
| Complexity_Proof "(('f, 'l) lab, 'v) strategy"
            "(('f, 'l) lab, 'v) rules"
            "(('f, 'l) lab, 'v) rules option"
            "(('f,'l)lab,'v)complexity_measure" complexity_class
            "(('f, 'l) lab, 'v) complexity_proof"
| DP_Proof "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) strategy"
  "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) rules"
  "('f, 'l, 'v) dp_proof"
| DP_Disproof "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) strategy"
  "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) dp_disproof"
| TRS_Disproof "(('f, 'l) lab, 'v) strategy"
               "(('f, 'l) lab, 'v) rules"
               "(('f, 'l) lab, 'v) trs_disproof"
| Rel_TRS_Disproof "(('f, 'l) lab, 'v) strategy"
                   "(('f, 'l) lab, 'v) rules"
                   "(('f, 'l) lab, 'v) rules"
                   "(('f, 'l) lab, 'v) reltrs_disproof"
| TRS_Confluence "(('f, 'l) lab, 'v) rules"
                 "('f, 'l, 'v) cr_proof"
| TRS_Non_Confluence "(('f, 'l) lab, 'v) rules"
                 "('f, 'l, 'v) ncr_proof"
| Completion_Proof "(('f, 'l) lab, 'v) equation list"
                   "(('f, 'l) lab, 'v) rules"
                   "('f, 'l, 'v) completion_proof"
| Equational_Proof "(('f, 'l) lab, 'v) equation list"
                   "(('f, 'l) lab, 'v) term \<times> (('f, 'l) lab, 'v) term"
                   "('f, 'l, 'v) equational_proof"
| Equational_Disproof "(('f, 'l) lab, 'v) equation list"
                   "(('f, 'l) lab, 'v) term \<times> (('f, 'l) lab, 'v) term"
                   "('f, 'l, 'v) equational_disproof"

fun input :: "('f, 'l, 'v) cert_problem \<Rightarrow> (('f, 'l) lab, 'v) input_problem" where
  "input (TRS_Proof Q R S_o _) = TRS Q R (rel_rules_of S_o)"
| "input (Complexity_Proof Q R S_o cm cc _) = Complexity_TRS Q R (rel_rules_of S_o) cm cc"
| "input (TRS_Disproof Q R _) = TRS Q R []"
| "input (Rel_TRS_Disproof Q R S _) = TRS Q R S"
| "input (DP_Proof P Pw Q R Rw _) = DP P Pw Q R Rw"
| "input (DP_Disproof P Q R _) = DP P [] Q [] R"
| "input (TRS_Confluence R _) = TRS No_Strategy R []"
| "input (TRS_Non_Confluence R _) = TRS No_Strategy R []"
| "input (Completion_Proof E R _) = EQS_TRS E R"
| "input (Equational_Proof E eq _) = EQS_EQ E eq"
| "input (Equational_Disproof E eq _) = EQS_EQ E eq"

fun desired_property :: "('f, 'l, 'v) cert_problem \<Rightarrow> bool" where
  "desired_property (TRS_Proof Q R S_o _ ) = SN_qrel (True,set (strategy_to_Q Q R), set R, set (rel_rules_of S_o))"
| "desired_property (Complexity_Proof Q R S_o cm cc _ ) = deriv_bound_measure_class (rel_qrstep (True,set (strategy_to_Q Q R), set R, set (rel_rules_of S_o))) cm cc"
| "desired_property (TRS_Disproof Q R _) = (\<not> SN (qrstep False (set (strategy_to_Q Q R)) (set R)))"
| "desired_property (Rel_TRS_Disproof Q R S _) = (\<not> SN_qrel (False,set (strategy_to_Q Q R), set R, set S))"
| "desired_property (DP_Proof P Pw Q R Rw _) = finite_dpp (True,True,set P, set Pw, set (strategy_to_Q Q (R @ Rw)), set R, set Rw)"
| "desired_property (DP_Disproof P Q R _) = infinite_dpp (False,set P, set (strategy_to_Q Q R), set R)"
| "desired_property (TRS_Confluence R _) = CR (rstep (set R))"
| "desired_property (TRS_Non_Confluence R _) = (\<not> CR (rstep (set R)))"
| "desired_property (Completion_Proof E R _) = completed_rewrite_system (set E) (set R)"
| "desired_property (Equational_Proof E eq _) = (eq \<in> eq_theory (set E))"
| "desired_property (Equational_Disproof E eq _) = (eq \<notin> eq_theory (set E))"

fun xml_tag :: "('f, 'l, 'v) cert_problem \<Rightarrow> string" where
  "xml_tag (TRS_Proof _ _ (Some _) _ ) = ''<relativeTerminationProof>''"
| "xml_tag (TRS_Proof _ _ None _ ) = ''<trsTerminationProof>''"
| "xml_tag (Complexity_Proof _ _ _ _ _ _ ) = ''<complexityProof>''"
| "xml_tag (TRS_Disproof _ _ _) = ''<trsNonterminationProof>''"
| "xml_tag (Rel_TRS_Disproof _ _ _ _) = ''<relativeNonterminationProof>''"
| "xml_tag (DP_Proof _ _ _ _ _ _) = ''<dpProof>''"
| "xml_tag (DP_Disproof _ _ _ _) = ''<dpNonterminationProof>''"
| "xml_tag (TRS_Confluence _ _) = ''<crProof>''"
| "xml_tag (TRS_Non_Confluence _ _) = ''<crDisproof>''"
| "xml_tag (Completion_Proof _ _ _) = ''<completionProof>''"
| "xml_tag (Equational_Proof _ _ _) = ''<equationalProof>''"
| "xml_tag (Equational_Disproof _ _ _) = ''<equationalDisproof>''"

fun cert_assms :: "bool \<Rightarrow> ('f, 'l, 'v) cert_problem \<Rightarrow> ('f, 'l, 'v) prob list" where
  "cert_assms a (TRS_Proof _ _ _ p) = assms a p"
| "cert_assms a (TRS_Confluence _ p) = cr_assms a p"
| "cert_assms a (TRS_Non_Confluence _ p) = ncr_assms a p"
| "cert_assms a (Completion_Proof _ _ p) = completion_assms a p"
| "cert_assms a (DP_Proof _ _ _ _ _ p) = map Inr (dp_assms a p)"
| "cert_assms a (Equational_Proof _ _ p) = equational_assms a p"
| "cert_assms a (Equational_Disproof _ _ p) = equational_dis_assms a p"
| "cert_assms _ _ = []"

lemma cert_assms_False[simp]: "cert_assms False p = []" by (cases p) simp_all

datatype cert_result = Certified string | Unsupported string | Error string

fun
  certify_cert_problem ::
    "bool \<Rightarrow> ('tp, ('f::{show,linorder}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> 
     ('f, label_type, string) cert_problem \<Rightarrow> cert_result"
where
  "certify_cert_problem a I J (TRS_Proof Q R None prf) =
    (case check_trs_proof a (shows_string ''1'') I J (tp_ops.mk I True (strategy_to_Q Q R) R []) prf of
      Inl err \<Rightarrow> Error ((''termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Proof Q R None prf)))"
| "certify_cert_problem a I J (TRS_Proof Q R (Some S) prf) = 
    (case check_trs_proof a (shows_string ''1'') I J (tp_ops.mk I True (strategy_to_Q Q R) R S) prf of
      Inl err \<Rightarrow> Error ((''relative termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Proof Q R (Some S) prf)))"
| "certify_cert_problem a I J (Complexity_Proof Q R S_o cm cc prf) = 
    (case check_complexity_proof (shows_string ''1'') I (tp_ops.mk I True (strategy_to_Q Q R) R (rel_rules_of S_o)) cm cc prf of
      Inl err \<Rightarrow> Error ((''complexity proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Complexity_Proof Q R S_o cm cc prf)))"
| "certify_cert_problem a I J (TRS_Disproof Q R prf) =
    (case check_trs_disproof (shows_string ''1'') I J (tp_ops.mk I False (strategy_to_Q Q R) R []) prf of
      Inl err \<Rightarrow> Error ((''nontermination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Disproof Q R prf)))"
| "certify_cert_problem a I J (Rel_TRS_Disproof Q R S prf) =
    (case check_reltrs_disproof (shows_string ''1'') I J (tp_ops.mk I False (strategy_to_Q Q R) R S) prf of
      Inl err \<Rightarrow>
        Error ((''relative nontermination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Rel_TRS_Disproof Q R S prf)))"
| "certify_cert_problem a I J (DP_Proof P Pw Q R Rw prf) =
    (case check_dp_proof_with_unlab a (shows_string ''1'') J (dpp_ops.mk J True True P Pw (strategy_to_Q Q (R @ Rw)) R Rw) prf of
      Inl err \<Rightarrow> Error ((''finiteness proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (DP_Proof P Pw Q R Rw prf)))"
| "certify_cert_problem a I J (DP_Disproof P Q R prf) =
    (case check_dp_disproof (shows_string ''1'') J (dpp_ops.mk J False False P [] (strategy_to_Q Q R) [] R) prf of
      Inl err \<Rightarrow> Error ((''infiniteness proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (DP_Disproof P Q R prf)))"
| "certify_cert_problem a I J (TRS_Confluence R prf) =
    (case check_cr_proof a (shows_string ''1'') I J R prf of
      Inl err \<Rightarrow> Error ((''confluence proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Confluence R prf)))"
| "certify_cert_problem a I J (TRS_Non_Confluence R prf) =
    (case check_ncr_proof a (shows_string ''1'') I J R prf of
      Inl err \<Rightarrow> Error ((''confluence disproof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Non_Confluence R prf)))"
| "certify_cert_problem a I J (Completion_Proof E R prf) =
    (case check_completion_proof a (shows_string ''1'') I J E R prf of
      Inl err \<Rightarrow> Error ((''completion proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Completion_Proof E R prf)))"
| "certify_cert_problem a I J (Equational_Proof E eq prf) =
    (case check_equational_proof a (shows_string ''1'') I J E eq prf of
      Inl err \<Rightarrow> Error ((''equational proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Equational_Proof E eq prf)))"
| "certify_cert_problem a I J (Equational_Disproof E eq prf) =
    (case check_equational_disproof a (shows_string ''1'') I J E eq prf of
      Inl err \<Rightarrow> Error ((''equational disproof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Equational_Disproof E eq prf)))"

lemma certify_cert_problem_with_assms_sound:
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and answer: "certify_cert_problem a I J cp = Certified info"
  shows "((\<forall>p\<in>set (cert_assms a cp). holds p) \<longrightarrow> desired_property cp) \<and> xml_tag cp = info"
proof -
  let ?i = "shows_string ''1''"
  interpret tp_spec I by (rule I)  
  show ?thesis
  proof (cases cp)
    case (TRS_Proof Q R sOpt prof)
    let ?tp = "tp_ops.mk I True (strategy_to_Q Q R) R"
    show ?thesis
    proof (cases sOpt)
      case None
      from answer have ok: "isOK (check_trs_proof a ?i I J (?tp []) prof) \<and> xml_tag cp = info" 
        unfolding TRS_Proof None
        by (cases "check_trs_proof a ?i I J (?tp []) prof") auto
      from check_trs_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
       and ok show ?thesis unfolding TRS_Proof None by simp
    next
      case (Some S)
      from answer have ok: "isOK (check_trs_proof a ?i I J (?tp S) prof) \<and> xml_tag cp = info"
        unfolding TRS_Proof Some
        by (cases "check_trs_proof a ?i I J (?tp S) prof", auto)
      from check_trs_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]] 
        and ok show ?thesis unfolding TRS_Proof Some by simp
    qed
  next
    case (Complexity_Proof Q R S_o cm cc prof)
    let ?Q = "strategy_to_Q Q R"
    let ?S = "rel_rules_of S_o"
    let ?tp = "tp_ops.mk I True ?Q R ?S"
    from answer have ok: "isOK (check_complexity_proof ?i I ?tp cm cc prof) \<and> xml_tag cp = info" 
      unfolding Complexity_Proof 
      by (cases "check_complexity_proof ?i I ?tp cm cc prof") auto
    from check_complexity_proof_sound[OF I ok[THEN conjunct1]]
      and ok show ?thesis unfolding Complexity_Proof by simp
  next
    case (TRS_Disproof Q R "prf")
    let ?tp = "tp_ops.mk I False (strategy_to_Q Q R) R []"
    from answer have ok: "isOK (check_trs_disproof ?i I J ?tp prf) \<and> xml_tag cp = info"
      unfolding TRS_Disproof
      by (cases "check_trs_disproof ?i I J ?tp prf") auto
    from check_trs_disproof_sound[OF I J ok[THEN conjunct1]]
      ok show ?thesis unfolding mk_simps TRS_Disproof by simp
  next
    case (Rel_TRS_Disproof Q R S "prf")
    let ?tp = "tp_ops.mk I False (strategy_to_Q Q R) R S"
    from answer have ok: "isOK (check_reltrs_disproof ?i I J ?tp prf) \<and> xml_tag cp = info"
      unfolding Rel_TRS_Disproof
      by (cases "check_reltrs_disproof ?i I J ?tp prf") auto
    from check_reltrs_disproof_sound[OF I J ok[THEN conjunct1]]
      ok show ?thesis unfolding mk_simps Rel_TRS_Disproof by simp
  next
    case (DP_Proof P Pw Q R Rw "prf")
    let ?dp = "dpp_ops.mk J True True P Pw (strategy_to_Q Q (R @ Rw)) R Rw"
    from answer have ok: "isOK (check_dp_proof_with_unlab a ?i J ?dp prf) \<and> xml_tag cp = info"
      unfolding DP_Proof
      by (cases "check_dp_proof_with_unlab a ?i J ?dp prf", auto)
    interpret J: dpp_spec J by (rule J)
    show ?thesis
    proof (intro conjI impI)
      assume ass: "\<forall> p \<in> set (cert_assms a cp). holds p"
      show "desired_property cp"
        unfolding DP_Proof desired_property.simps
      proof (rule check_dp_proof_with_unlab_dp_assms[OF J _ ok[THEN conjunct1], unfolded 
            J.dpp_spec_sound J.mk_simps cert_assms.simps])
        from ass
        show "\<forall> p \<in> set (dp_assms a prf). finite_dppLAll p"
          unfolding DP_Proof cert_assms.simps
          by (induct "dp_assms a prf", auto)
      qed
    qed (insert ok, auto)
  next
    case (DP_Disproof P Q R "prf")
    let ?dp = "dpp_ops.mk J False False P [] (strategy_to_Q Q R) [] R"
    from answer have ok: "isOK (check_dp_disproof ?i J ?dp prf) \<and> xml_tag cp = info"
      unfolding DP_Disproof
      by (cases "check_dp_disproof ?i J ?dp prf", auto)
    interpret J: dpp_spec J by (rule J)
    show ?thesis
    proof (intro conjI impI)
      show "desired_property cp"
        unfolding DP_Disproof desired_property.simps
        using check_dp_disproof[OF J infinite_lab ok[THEN conjunct1], unfolded 
            J.dpp_spec_sound J.mk_simps]
        by auto
    qed (insert ok, auto)
  next
    case (TRS_Confluence R "prf")
    from answer have ok: "isOK (check_cr_proof a ?i I J R prf) \<and> xml_tag cp = info"
      unfolding TRS_Confluence
      by (cases "check_cr_proof a ?i I J R prf", auto)
    note res = check_cr_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding TRS_Confluence by auto
  next
    case (TRS_Non_Confluence R "prf")
    from answer have ok: "isOK (check_ncr_proof a ?i I J R prf) \<and> xml_tag cp = info"
      unfolding TRS_Non_Confluence
      by (cases "check_ncr_proof a ?i I J R prf", auto)
    note res = check_ncr_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding TRS_Non_Confluence by auto
  next
    case (Completion_Proof E R "prf")
    from answer have ok: "isOK (check_completion_proof a ?i I J E R prf) \<and> xml_tag cp = info"
      unfolding Completion_Proof
      by (cases "check_completion_proof a ?i I J E R prf", auto)
    note res = check_completion_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Completion_Proof by auto
  next
    case (Equational_Proof E eq "prf")
    from answer have ok: "isOK (check_equational_proof a ?i I J E eq prf) \<and> xml_tag cp = info"
      unfolding Equational_Proof by (cases "check_equational_proof a ?i I J E eq prf") auto
    note res = check_equational_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Equational_Proof by auto
  next
    case (Equational_Disproof E eq "prf")
    from answer have ok: "isOK (check_equational_disproof a ?i I J E eq prf) \<and> xml_tag cp = info"
      unfolding Equational_Disproof by (cases "check_equational_disproof a ?i I J E eq prf") auto
    note res = check_equational_disproof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Equational_Disproof by auto
  qed
qed

lemma certify_cert_problem_sound:
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and answer: "certify_cert_problem False I J cp = Certified info"
  shows "desired_property cp \<and> xml_tag cp = info"
  using certify_cert_problem_with_assms_sound[OF I J answer] by simp

end
