theory Loga
imports "../TermFun"
begin

section {* Example - The Binary Logarithm *}

text {* Compute the quotient of dividing a natural number by two. *}
fun half where
  "half 0 = 0"
| "half (Suc 0) = 0"
| "half (Suc (Suc n)) = Suc (half n)"

text {* Compute the binary logarithm by successive bisection. *}
function log where
  "log n = (case half n of 0 \<Rightarrow> 0 | Suc m \<Rightarrow> Suc (log (Suc m)))"
  by pat_completeness auto

text {* Generate first-order term rewrite systems for the defining equations
of @{const "half"} and @{const "log"}. The @{text "no_sim"} attribute in the
latter case is needed to prevent the system from proving the \emph{simulation lemma}
for @{const "log"} (which is only possible after termination has been established).
(Consult the paper for more details.) *}
define_trs half.simps
define_trs (no_sim) log.psimps

text {* The resulting TRSs are available under the names @{const "half.trs"} and
@{const "log.trs"}, having the definitions @{thm[source] half.trs_def} and
@{thm[source] log.trs_def}, respectively. *}
term half.trs
term log.trs
thm half.trs_def log.trs_def

text {*
Finally, we prove termination of @{const "log"}. The initial obligation is
to show that all inputs are in the domaon of @{const "log"} (and hence, the function
is defined on all inputs). *}
termination log
  txt {*
  We first reduce this obligation to termination of @{const "log"}s TRS
  representation (the main content of our paper).
  *}
  apply (trs_reduce log)
  txt {* Before we export the TRS, we transform it into a finite list of explicit rules. *}
  apply (simp only: all_trs_defs set_list_conv)
  txt {* Switch from innermost termination to termination by now *}
  apply (rule SN_subset[OF _ qrstep_NF_anti_mono, of _ "set []"])
  txt {* The result can directly be written to an external file (as XML). *}
  export_termination_problem "log.xml"
  apply (ceta_import_proof "proofs/log.proof.xml")
  apply ceta_eval
  apply auto
done

end
