theory Mktree
imports "../TermFun"
begin

section{*Definition of Huffman's algorithm*}

text {* taken from Nipkow's Isabelle presentation at TPHOLs'08 *}

datatype 'a tree = Leaf nat 'a | Node nat "('a tree)" "('a tree)"

primrec "value" where
"value (Leaf n _) = n" |
"value (Node n _ _) = n"

definition merge where
  "merge t1 t2 = Node (value t1 + value t2) t1 t2"

fun ins where
"ins u (t#ts) = (if value u \<le> value t then u # t # ts else t # ins u ts)" |
"ins u [] = [u]"

function (sequential) mkTree where
"mkTree(t1#t2#ts) = mkTree(ins (merge t1 t2) ts)" |
"mkTree[t] = t"
by pat_completeness auto




setup {* Term_Fun.register_datatypes [@{type_name tree}] *}
define_trs value.simps
define_trs merge_def
define_trs ins.simps

 (* here we can get away with underspecification,
since we do not need the simulation lemma *)
define_trs (no_sim) mkTree.psimps

thm mkTree.trs_def
termination mkTree
apply (trs_reduce mkTree)
apply (simp only: all_trs_defs set_list_conv)
apply (rule SN_subset[OF _ qrstep_NF_anti_mono, of _ "set []"])
export_termination_problem "mkTree.xml"
apply (ceta_import_proof "proofs/mkTree.proof.xml")
apply ceta_eval
apply auto
done

end