theory TermFun_Setup
imports "../Ceta_Checker"
keywords "define_trs" :: thy_decl and "define_dps" :: thy_decl and "export_termination_problem" :: diag
uses
  ("ML/term.ML")
  ("ML/basics.ML") 
  ("ML/call_tree.ML") 
  ("ML/trs_data.ML")
  ("ML/tactics.ML")
  ("ML/embed.ML")
  ("ML/term_fun.ML")
  ("ML/xml.ML")
begin

section {* Preliminaries *}

(*<*)
consts
  emb_const :: "'a \<Rightarrow> (string, string) term list \<Rightarrow> (string, string) term"
  (infix "<>" 100)

setup {*
let
  fun quote (Const (@{const_name emb_const}, _) $ Const (n, _)) =
      SOME (@{term "Fun :: string \<Rightarrow> (string,string) term list \<Rightarrow> (string,string) term"} $ HOLogic.mk_string n)
    | quote _ = NONE

  fun unquote (Const (@{const_name Term.Fun}, _) $ n) =
      ((SOME (Const (@{const_name emb_const}, dummyT)
        $ Const(HOLogic.dest_string n, dummyT)))
      handle TERM _ => NONE)
    | unquote _ = NONE

  fun rew f ts ctxt =
    let 
      val thy = Proof_Context.theory_of ctxt
      val ts' = map (Pattern.rewrite_term thy [] [f]) ts
    in
      if forall (op aconv) (ts ~~ ts') then NONE
      else SOME (ts', ctxt)
    end
in
   Context.theory_map 
    (Syntax_Phases.term_check' ~1 "quote_funs" (rew quote)
    #> Syntax_Phases.term_uncheck' 1 "quote_funs" (rew unquote)) 
end
*}
(*>*)


lemma length_eqI: "length xs = length ys \<Longrightarrow> length (x#xs) = length (y#ys)"
  by auto

lemma pairwiseI1: "\<forall>i<length []. P i"
  and pairwiseI2: "(x, y) \<in> R
  \<Longrightarrow> \<forall>i<length xs. (xs ! i, ys ! i) \<in> R
  \<Longrightarrow> \<forall>i<length (x#xs). ((x#xs) ! i, (y#ys) ! i) \<in> R"
  by auto (case_tac i, auto)

lemma list_memberI1: "x \<in> set (x # xs)"
  and list_memberI2: "y \<in> set xs \<Longrightarrow> y \<in> set (x # xs)"
  by auto

lemma fixes f shows inv_imageI: "(f a, f b) \<in> R \<Longrightarrow> (a, b) \<in> inv_image R f"
  by auto

lemmas into_CC =  subsetD[OF subset_ctxt_closure]

abbreviation (input) def_prefix :: "char" where "def_prefix \<equiv> CHR ''d''"
abbreviation (input) con_prefix :: "char" where "con_prefix \<equiv> CHR ''c''"

abbreviation (input) tf_Defs :: "string set" where "tf_Defs \<equiv> {f . f \<noteq> [] \<and> hd f = def_prefix}"
definition tf_Q :: "(string,string)terms" where "tf_Q \<equiv> {Fun f ts | f ts. f \<in> tf_Defs}"
abbreviation (input) tf_QL :: "(string,string)term list" where "tf_QL \<equiv> []"

definition map_constructor :: "(string,string)term \<Rightarrow> (string,string)term"
  where "map_constructor \<equiv> map_term (Cons con_prefix)"

lemma map_constructor_Fun[simp]: "map_constructor (Fun f ts) = Fun (Cons con_prefix f) (map map_constructor ts)"
  unfolding map_constructor_def by simp

definition check_constructor :: "(string,string)term \<Rightarrow> bool"
  where "check_constructor \<equiv> \<lambda> t. funs_term t \<inter> tf_Defs = {}"

lemma check_constructor_map_constructor[simp]: "check_constructor (map_constructor t)"
  unfolding check_constructor_def map_constructor_def
  by (induction t, auto)

lemma check_constructor_fun[simp]: "check_constructor (Fun f ts) = ((f = [] \<or> hd f \<noteq> def_prefix) \<and> Ball (set ts) check_constructor)"
  unfolding check_constructor_def by auto

lemmas check_constructor_simps = check_constructor_map_constructor check_constructor_fun

lemma check_constructor_NF: assumes check: "check_constructor t"
  shows "t \<in> NF_terms tf_Q"
proof
  fix C l \<sigma>
  assume "l \<in> tf_Q" and t: "t = C \<langle> l \<cdot> \<sigma> \<rangle>" 
  from this[unfolded tf_Q_def] obtain f ls where l: "l = Fun f ls" and f: "f \<in> tf_Defs" by auto
  have "(f,length ls) \<in> funas_term t" unfolding t l by simp
  hence ft: "f \<in> funs_term t" unfolding funs_term_funas_term by force
  with check[unfolded check_constructor_def] f show False by auto
qed

abbreviation dp_step :: "(string,string)trs \<Rightarrow> (string,string)trs"
  where "dp_step P \<equiv> rqrstep True tf_Q P \<inter> NF_terms tf_Q \<times> UNIV"    

lemma tf_Q_rqrstepI: assumes lr: "(l,r) \<in> R"
  and no_defs: "Ball (set (args (l \<cdot> \<sigma>))) check_constructor \<and> Ball (set (map \<sigma> (vars_rule_list (l,r)))) check_constructor"
  shows "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> rqrstep True tf_Q R"
  by (rule rqrstepI[OF _ lr refl refl], unfold NF_terms_args_conv[symmetric] NF_subst_def,
  insert no_defs, auto intro: check_constructor_NF)

lemma tf_Q_qrstepI: assumes lr: "(l,r) \<in> R"
  and no_defs: "Ball (set (args (l \<cdot> \<sigma>))) check_constructor \<and> Ball (set (map \<sigma> (vars_rule_list (l,r)))) check_constructor"
  shows "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> qrstep True tf_Q R"
  using tf_Q_rqrstepI[OF lr no_defs] unfolding qrstep_iff_rqrstep_or_nrqrstep by auto

lemma dp_stepI: assumes lr: "(l,r) \<in> P"
  and no_defs: "check_constructor (l \<cdot> \<sigma>) \<and> Ball (set (map \<sigma> (vars_rule_list (l,r)))) check_constructor"
    (is "?a \<and> ?b")
  shows "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> dp_step P"
proof -
  from no_defs have "Ball (set (args (l \<cdot> \<sigma>))) check_constructor" (is ?c)
    by (cases "l \<cdot> \<sigma>", auto)
  with no_defs have "?c \<and> ?b" and "?a" by auto
  from tf_Q_rqrstepI[OF lr this(1)] check_constructor_NF[OF `?a`]
  show ?thesis by simp
qed

lemmas stepI_simps = Term.subst_apply_term.simps
               Trs_Impl.vars_rule_list_def 
               Trs_Impl.vars_rule_list'_def 
               Term_Impl.vars_term_list'.simps
               Product_Type.fst_conv 
               Product_Type.snd_conv
               List.map.simps 
               Fun.fun_upd_apply
               HOL.if_True
               HOL.if_False
               List.list.inject
               List.list.distinct
               String.char.inject
               String.nibble.eq.simps
               String.nibble.distinct
               HOL.refl
               List.Ball_set
               List.remdups.simps
               Term.args.simps
               List.list_all_simps
               List.append.simps
               TermFun_Setup.check_constructor_simps
               List.hd.simps
               List.list.simps
               List.concat.simps
               HOL.simp_thms
               List.set.simps
               Set.empty_iff
               Set.insert_iff
               Set.ball_simps

abbreviation "sub_steps R \<equiv> (supt \<union> qrstep True tf_Q R)^*"

definition "dpr_step P R \<equiv> relto (dp_step P) (qrstep True tf_Q R)"

abbreviation "dpr_steps P R \<equiv> (dp_step P \<union> qrstep True tf_Q R)^*"

lemma qterm_relI_initial:
  "(x, y) \<in> qrstep True tf_Q R \<Longrightarrow> (y, z) \<in> sub_steps R \<Longrightarrow> (x,z) \<in> (supt \<union> qrstep True tf_Q R)^+"
  by (rule rtrancl_into_trancl2) (rule UnI2)

lemma dp_qterm_relI_initial:
  "(x, y) \<in> dp_step P \<Longrightarrow> (y, z) \<in> dpr_steps P R \<Longrightarrow> (x,z) \<in> (dpr_step P R)^+"
  unfolding relto_trancl_conv dpr_step_def by blast

lemma dp_SN_dpr_step_trancl: "SN (dpr_step P R) \<Longrightarrow> SN ((dpr_step P R)^+)"
  by (rule SN_imp_SN_trancl)

lemma SN_supt_qrstep_trancl: "SN (qrstep True Q R) \<Longrightarrow> SN ((supt \<union> qrstep True Q R)^+)"
  using SN_imp_SN_trancl[OF SN_imp_SN_union_supt[OF _ ctxt_closed_qrstep], of True Q R] 
  unfolding Un_commute[of supt "qrstep True Q R"] .

lemma qterm_relI_step:
  "(x, y) \<in> qrstep True tf_Q R \<Longrightarrow> (y, z) \<in> sub_steps R \<Longrightarrow> (x,z) \<in> sub_steps R"
  by (rule converse_rtrancl_into_rtrancl) (rule UnI2)

lemma qterm_relI_steps:
  "(x, y) \<in> (qrstep True tf_Q R)^* \<Longrightarrow> (y, z) \<in> sub_steps R \<Longrightarrow> (x,z) \<in> sub_steps R"
  apply (rule rtrancl_trans)
  apply (rule subsetD[OF rtrancl_mono])
  by (rule Un_upper2)

lemma dp_qterm_relI_steps:
  assumes xy: "(x, y) \<in> (qrstep True tf_Q R)^*"
  and yz: "(y, z) \<in> dpr_steps P R"
  shows "(x,z) \<in> dpr_steps P R"
proof -
  have "(x,y) \<in> dpr_steps P R"
    by (rule set_mp[OF rtrancl_mono xy], auto)
  with yz show ?thesis by auto
qed

lemma qterm_relI_sub:
  "(x, y) \<in> supt \<Longrightarrow> (y, z) \<in> sub_steps R \<Longrightarrow> (x,z) \<in> sub_steps R"
  by (rule converse_rtrancl_into_rtrancl) (rule UnI1)

class trs_rep =
  fixes embed :: "'a \<Rightarrow> (string, string) term" ("\<lceil>_\<rceil>")

definition dummy_term :: "(string, string) term" where
  "dummy_term \<equiv> Fun ''dummy_term'' []"
notation (output) dummy_term ("\<bullet>")


use "ML/term.ML"
use "ML/basics.ML"
use "ML/call_tree.ML"
use "ML/trs_data.ML"
use "ML/tactics.ML"
use "ML/embed.ML"
use "ML/xml.ML"
use "ML/term_fun.ML"
setup Term_Fun.setup

(* Example proof to test startup-time of ceta_eval.
   very fast, if code_reflect and ceta_checker_def[code del] are used *)
lemma "SN (qrstep True (set tf_QL) (set ([]::(string, string) rules)))"
  apply (rule ceta_checker[of _ _ "RisEmpty::(string, label_type, string) trs_proof"])
  apply ceta_eval
done


end